(* This file is part of MLGame (OCaml Game System).

   MLGame is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2 of the License, or (at your option) any later
   version.

   MLGame is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
   FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
   details.

   You should have received a copy of the GNU General Public License along
   with MLGame; if not, write to the Free Software Foundation, Inc., 59 Temple
   Place, Suite 330, Boston, MA 02111-1307 USA *)

(** Parser *)

(** Parser raises Error if it is unable to parse the given string. *)
exception Error of string;;

(** A type of information that can be processed by the parser. It knows which
   values are acceptable and sometimes knows how to complete a word. Some
   such types will also know how to display themselves in the menu. *) 
type 'a t = {
  lex: (Lexing.lexbuf -> 'a);
  deparse: ('a -> string);
  complete: (Lexing.lexbuf -> string list);
  accept: ('a -> 'a);
  desc: string;
};;

(** Simple command and variable params *)
val int : ?max:int -> ?min:int -> unit -> int t;;
val float : unit -> float t;;
val unit : unit t;;
val bool : bool t;;
val string : string t;;

(** For optional parameters in commands *)
val option : 'a t -> 'a option t;;

(** For functions with more arguments *)
val pair : 'a t -> 'b t -> ('a * 'b) t;;
val triple : 'a t -> 'b t -> 'c t -> ('a * 'b * 'c) t;;
val quadruple : 'a t -> 'b t -> 'c t -> 'd t -> ('a * 'b * 'c * 'd) t;;
val pentuple : 'a t -> 'b t -> 'c t -> 'd t -> 'e t -> ('a * 'b * 'c * 'd * 'e) t;;
				    
(** If we want to accept only some arguments of the given type *)
val accept_only : 'a list -> 'a t -> 'a t;;

(** Adds a callback to a change. The callback will be called with
   given 'a whenever the parser variable of this type is changed *)
val callback : ('a -> unit) -> 'a t -> 'a t;;

(** For commands or variables taking commands as parameters. The command
   can be a simple command (not taking arguments) or a one in \{such\} 
   brackets with params inside. Completes to a command and returns the 
   parsed command as argument 
   
   For example, the parser command "alias" takes a shortcut and a command
      to call.
   A parser variable (that is a one which we can get/set from the parser)
     can also be of command type.
   *)
val command : (unit -> unit) t;;

(** For commands that can take variable names as parameters. *)
val variable : string t;;

(** [add_command name type description handler] adds the command to the list
   of commands. Examples of usage in MLGame:

   Parser.add_command "exec" Parser.string "executes the given script" exec

   Parser.add_command "bind_global" (Parser.pair Key.parser Parser.command)
     "Bind_global binds..." my_function;;

   Parser.add_command "disconnect" Net.local_player_parser "This fun.." fun;;*)
val add_command : string -> 'a t -> string -> ('a -> unit) -> unit;;

(** Adds a parser variable with given name and of given type with given
   initial value. The reference to the variable is returned. The 
   variable is accessible both from the parser by its name or from
   the program by returned reference. *)
val add_variable : string -> 'a t -> 'a -> 'a ref;;

(** Gets the value of a parser variable in string format, as would user
   do it through the console. @raise Not_found if there is no such
   variable *)
val get : string -> string;;

(** Sets the variable, like if the user would have written the set command
   on the console *)
val set : string -> string -> unit;;

(** [parse expression] Parses a given string and returns a simple function
   that executes the parsed [expression]. Thanks to second unit argument
   you can parse the command before really executing it.
   @raise Error if something went wrong. *)
val parse : string -> unit -> unit;;

(** [available_completions str] Returns all possible available completions 
   of [str]. It completes command names, their arguments, variable names etc... *)
val available_completions : string -> string list;;

(** [create_log name prefix init] creates a debug logger, and returns a function
  used to log with this logger. Takes a prefix [prefix] for all logs that are logged
  with this command, a [name] which creates a parser variable to turn off and on
  displaying of this debug info and [init] which is the initial state of logging
  *)
val create_log : string -> string -> bool -> (string -> unit);;

val sextuple :
  'a t ->
  'b t -> 'c t -> 'd t -> 'e t -> 'f t -> ('a * 'b * 'c * 'd * 'e * 'f) t;;
val septuple :
  'a t -> 'b t -> 'c t ->
  'd t -> 'e t -> 'f t -> 'g t -> ('a * 'b * 'c * 'd * 'e * 'f * 'g) t;;
val octuple :
  'a t -> 'b t -> 'c t -> 'd t ->
  'e t -> 'f t -> 'g t -> 'h t -> ('a * 'b * 'c * 'd * 'e * 'f * 'g * 'h) t;;
val nonuple :
  'a t -> 'b t -> 'c t -> 'd t -> 'e t -> 'f t ->
  'g t -> 'h t -> 'i t -> ('a * 'b * 'c * 'd * 'e * 'f * 'g * 'h * 'i) t;;
