
(****************** TYPES *******************)

type state = {
  (* status *)
  mutable pos : Types.fpos; (* position *)
  mutable vel : Types.vel; (* velocity *)
  mutable health : int; (* health points (from 1 to max_health) *)
  max_health : int; (* maximum health points *)
  mutable lives_c : int; (* number of lives left *)
  mutable kills_c : int; (* number of kills *)
  mutable weapon : t Weapon.t; (* currently selected weapon *)
  mutable available_weapons : t Weapon.t list; (* all weapons owned by player *)
  (* subobjects *)
  crosshair : Crosshair.t; (* the crosshair (sight) *)
  hook : Hook.t; (* the hook (robe) *)
  (* players set *)
  pl_no : int; (* player number *)
  players : set; (* set of all players in which we are *)
  (* temporary *)
  mutable flying_robe : float; (* When the change+jump combination is
      pressed and the robe is already thrown the value is 0, in this case
      the robe should not be hiden and thrown again. When the combination is
      released it counts time from this moment, so immediately after
      releasing the combination jumping will not be possible. *)
  mutable last_weapon_ch : float; (* time from last weapon change *)
}
and t = state Sprite.t
and set = (int, t) Hashtbl.t * state Sprite.set * Arena.t;;

type shoot_params = t Weapon.shoot_params;;

type status = {
  health100 : int;
  ammo100 : int;
  is_reloading : bool;
  lives : int;
  kills : int;
};;

(******************* GETTERS **********************)

let get_status pl = 
  let data = Sprite.get_data pl in
  { health100 = data.health * 100 / data.max_health;
    ammo100 = Weapon.get_ammo100 data.weapon;
    is_reloading = Weapon.is_reloading data.weapon;
    lives = data.lives_c;
    kills = data.kills_c
  }
;;

let get_pos = Sprite.get_pos;;
let get_angle pl = Crosshair.get_angle (Sprite.get_data pl).crosshair;;

let get_direction sprite =
  match Sprite.get_animation_frame sprite with 
    0 | 2 -> Types.Left | 1 | 3 -> Types.Right | _ -> assert false
;;
(*************** SET OF PLAYERS *****************)

let create_set arena = 
  let ht = Hashtbl.create 10 in
  let set = Sprite.Set.create 50 in
  Sprite.Set.set_collision set 32 32;
  Arena.set_boundary arena set;
  ht, set, arena
;;

let by_num (ht, _, _) = Hashtbl.find ht;;

let add (ht, _, _) = Hashtbl.replace ht;;

let delete pl = 
  let data = Sprite.get_data pl in
  let ht, _, _ = data.players in
  Hashtbl.remove ht data.pl_no;
  Sprite.delete pl
;;

let iter f (ht, _, _) = Hashtbl.iter (fun _ -> f) ht;;

(****************** HELPER FUNCTIONS ***********************)

let to_int (x, y) = (int_of_float x, int_of_float y);;

let can_go sprite pos =
  Sprite.move_to sprite (to_int pos);
  let _, _, arena = (Sprite.get_data sprite).players in
  Sprite.in_bounds sprite && Sprite.collides_with sprite true (Arena.get_sprite_set arena) = []
;;

let rec set_random_pos sprite = 
  let data = Sprite.get_data sprite in
  let _, _, arena = data.players in
  let _, _, x1, y1 = Arena.get_boundary arena in
  let x, y = float_of_int (Random.int x1), float_of_int (Random.int y1) in
  if can_go sprite (x, y) then
    data.pos <- x, y
  else
    set_random_pos sprite
;;

let rec find_weapon sprite = 
  let data = Sprite.get_data sprite in
  let rec find n = function 
      h :: t -> if h == data.weapon then n else find (n + 1) t
    | _ -> assert false
  in
  find 0 data.available_weapons
;;

(********************************* CREATE *******************)

let img = Video.load_image "Data/worms.png";;

let create (_, set, arena as players) pl_no weapons =
  let state =
    {pos = (0., 0.);
     vel = (0., 0.); 
     crosshair = Crosshair.create ();
     pl_no = pl_no;
     health = 200;
     max_health = 200;
     lives_c = !Constants.lives;
     kills_c = 0;
     weapon = List.hd weapons; 
     available_weapons = weapons;
     hook = Hook.create arena;
     players = players;
     last_weapon_ch = 0.;
     flying_robe = Constants.no_jump_after_robe;
   } in
  let sprite = 
    Sprite.create ~set:set ~center_pos:(11, Constants.player_center_y) ~width:24 state img (0, 0) 
  in
  if pl_no < 0 then Sprite.set_animation_frame sprite 2 else ();
  add players pl_no sprite;
  set_random_pos sprite
;;

(************************** MOVING ************************)

(* throw the rope *)
let release_rope sprite =
  let data = (Sprite.get_data sprite) in
  let (cx, cy) = Crosshair.get_pos data.crosshair in
  let (px, py) = data.pos in
  let (dx, dy) = float_of_int cx -. px, float_of_int cy -. py in
  let (vx, vy) = !Constants.rope_speed *. dx, !Constants.rope_speed *. dy in
  Hook.release data.hook (to_int data.pos) (vx, vy) 
;;

(* if we are staying on the ground *)
let on_ground sprite (px, py) =
  not (can_go sprite (px, py +. 1.))
;;

(* do the move, taking into account pressed keys *)
let move sprite =
  let data = Sprite.get_data sprite in
  let can_go = can_go sprite in
  let x, y = data.pos in
  assert (can_go data.pos);

  (* calculate how much we should go up / down after moving horizontaly *)
  (* climb positive -> falling down / negative - climbing up *)
  let rec climb_f (x, y) skill = 
    if (can_go (x, y +. skill)) || skill < -. !Constants.climbing_skill then skill
    else climb_f (x, y) (skill -. 0.5) 
  in

  (* set the proper direction of the worm *)
  let old_frame = Sprite.get_animation_frame sprite in
  let dir = if Controls.get_change data.pl_no then 0 else Controls.get_vx data.pl_no in
  begin match dir with
  | 1 -> Sprite.set_animation_frame sprite (if data.pl_no < 0 then 3 else 1)
  | -1 -> Sprite.set_animation_frame sprite (if data.pl_no < 0 then 2 else 0)
  | _ -> ()
  end;
  (* sometimes rotate is impossible (collision), try to go up, eventualy don't rotate *)
  let climb = climb_f (x, y) 0. in
  let climb = if climb < -. !Constants.climbing_skill then 0. else climb in
  let y = y +. climb in
  if not (can_go (x, y)) then Sprite.set_animation_frame sprite old_frame;

  (* calculate the new velocity (after gravity and rope force) *)
  let vx, vy = data.vel in
  let rvx, rvy = Hook.rope_force data.hook (to_int (x, y)) in
  let vx, vy = vx +. !Arena.delay *. rvx, vy +. !Arena.delay *. (rvy +. !Constants.weight) in
  let dy = !Arena.delay *. vy in

  let flying () = 
    (* when in the air *)
    let dfx = !Arena.delay *. (vx +. float_of_int dir *. !Constants.flight_steering_speed) in
    let y, vy = if can_go (x, y +. dy) then y +. dy, vy else y, 0. in
    let x, vx = if can_go (x +. dfx, y) then x +. dfx, vx else x, 0. in
    data.pos <- (x, y);
    data.vel <- (vx *. (1. -. !Arena.delay *. !Constants.air_viscosity), 
      vy *. (1. -. !Arena.delay *. !Constants.air_viscosity))
  in

  if not (on_ground sprite (x, y)) then
    flying ()
  else begin
    let dgx = !Arena.delay *. (vx +. float_of_int dir *. !Constants.move_speed) in
    (* calculate how much we should go up / down after moving horizontaly *)
    (* climb positive -> falling down / negative - climbing up *)
    let climb = climb_f (x +. dgx, y) !Constants.climbing_skill in
    if climb < -. !Constants.climbing_skill || (* can not climb *)
      (dy < 0. && dy <= climb) (* can go higher using velocity (rope force) than by climbing *)
    then 
      flying ()
    else begin
      data.pos <- (x +. dgx, y +. climb);
      data.vel <- (vx *. (1. -. !Arena.delay *. !Constants.resistance), 0.);
    end
  end;

  (* jumping *)
  let (vx, vy) = data.vel in
  let vy = if Controls.get_jump data.pl_no && not (Controls.get_change data.pl_no) 
    && on_ground sprite data.pos && data.flying_robe >= Constants.no_jump_after_robe then 
      vy -. !Constants.jump_speed else vy in
  data.vel <- (vx, vy);
  
  (* move the sprite *)
  Sprite.move_to sprite (to_int data.pos);
;;

let process sprite =
  let data = Sprite.get_data sprite in
  (* moving *)
  move sprite;
  Crosshair.process data.crosshair data.pl_no (get_direction sprite) (get_pos sprite);
  (* robe *)
  Hook.process data.hook;
  if Controls.get_jump data.pl_no && not (Controls.get_change data.pl_no) 
    && data.flying_robe >= Constants.no_jump_after_robe then
      Hook.hide data.hook;
  if Controls.get_change data.pl_no && Controls.get_jump data.pl_no then begin
    if data.flying_robe > 0. then
      release_rope sprite;
    data.flying_robe <- 0.
  end else
    data.flying_robe <- data.flying_robe +. !Arena.delay;
  (* reloading weapon / shooting *)
  Weapon.process data.weapon;
  if Controls.get_fire data.pl_no then
    Weapon.fire data.weapon {
      Weapon.player = Some sprite;
      Weapon.pos = get_pos sprite;
      Weapon.vel_mod = data.vel;
      Weapon.angle = Crosshair.get_angle data.crosshair;
      Weapon.dir = get_direction sprite;
    };
  (* weapon change *)
  let ch = Controls.get_vx data.pl_no in
  if Controls.get_change data.pl_no && not (Controls.get_jump data.pl_no) && ch <> 0 then begin
    data.last_weapon_ch <- data.last_weapon_ch +. !Arena.delay;
    if data.last_weapon_ch > Constants.weapon_ch_time then begin
      let len = List.length data.available_weapons in
      let n = find_weapon sprite in
      data.weapon <- List.nth data.available_weapons ((len + n - ch) mod len);
      data.last_weapon_ch <- 0.;
    end
  end else 
    data.last_weapon_ch <- Constants.weapon_ch_time
;;

(******************** DRAWING ****************************)

let font_white = Font.load "Data/fixed12s.png";;(*"Data/font.png";;*)
            
let draw sprite cntx = 
  let data = Sprite.get_data sprite in
  Sprite.draw ~on:cntx sprite;
  Crosshair.draw data.crosshair cntx;
  Hook.draw cntx data.hook (Sprite.get_pos sprite);
  let x, y = Sprite.get_pos sprite in
  if Controls.get_change data.pl_no && not (Controls.get_jump data.pl_no) then
    let name = Weapon.get_name data.weapon in
    let width = Font.width font_white name in
    Font.draw_string ~on:cntx font_white (x - width / 2) (y - 30) name 
;;

(**************** OTHER ********************)

let collides_with (_, set, _) sprite = Sprite.collides_with sprite true set;;

let injure sprite force by_who = 
  let data = Sprite.get_data sprite in
  data.health <- min data.max_health (data.health - force);
  if data.health <= 0 then begin (* death *)
    data.health <- data.max_health;
    data.lives_c <- data.lives_c - 1;
    Hook.hide data.hook;
    set_random_pos sprite;
    begin match by_who with
        None -> ()
      | Some killer -> if killer != sprite then
            let d2 = Sprite.get_data killer in
            d2.kills_c <- d2.kills_c + 1
    end;
    if data.lives_c = 0 then begin
      iter (fun pl -> 
      	let dat2 = Sprite.get_data pl in
      	print_string ("Player " ^ string_of_int dat2.pl_no ^ " - lives: " ^ 
      	  string_of_int dat2.lives_c ^ ", kills: " ^ string_of_int dat2.kills_c ^ "\n")
      ) data.players;
      raise Input.Quit
    end
  end
;;

let add_weapon sprite weapon =
  let data = Sprite.get_data sprite in
  let n = find_weapon sprite in
  let rec replace cur = function 
      h :: t -> 
        if cur = n then begin
          assert (h == data.weapon);
          weapon :: t
        end else h :: replace (cur + 1) t
    | _ -> assert false
  in
  assert (n < List.length data.available_weapons);
  data.available_weapons <- replace 0 data.available_weapons;
  data.weapon <- weapon
;;


(*
let rec find_min_free hashtbl i =
  if (Hashtbl.mem hashtbl i) then find_min_free hashtbl (i - 1) else i
;;

let create_ai () =
  let pl_no = find_min_free players (-1) in
  create pl_no
;;
  *)
(*Parser.add_command "comp" (Parser.unit) "" create_ai;;*)
