
type desc = {
  weight : float;
  air_viscosity : float;
  initial_vel : float;
  velocity_vary : float;
  life_time : float;
  bounce_res : float;
  img : Video.t;
  no_frames : int;
};;

let std_desc = {
  weight = 0.;
  air_viscosity = 0.;
  initial_vel = 300.;
  velocity_vary = 0.;
  life_time = 100.;
  bounce_res = 0.;
  img = Video.load_image "Data/fire.png";
  no_frames = 1;
};;

let to_int (x, y) = (int_of_float x, int_of_float y);;

class virtual t set desc (params : Player.shoot_params) = 
  let spr_w, spr_h = desc.img.Video.rect.Sdlvideo.r_w, desc.img.Video.rect.Sdlvideo.r_h in
  let spr_w = spr_w / desc.no_frames in
  object (self)
    inherit Bullet.t set as super

    val mutable sprite = 
      Sprite.create ~width:spr_w ~center_pos:(spr_w / 2, spr_h / 2) () desc.img 
        params.Weapon.pos

    (* we can move at most max_step at once *)
    val max_step = float_of_int (min spr_w spr_h)

    (* when initial is set then collision with the owner player will not count *)
    val mutable initial = true

    val mutable vel = 
      let vx, vy = params.Weapon.vel_mod in
      vx +. float_of_int (Types.dir_to_int params.Weapon.dir) *. 
          desc.initial_vel *. cos params.Weapon.angle +. Random.float desc.velocity_vary, 
        vy +. desc.initial_vel *. sin params.Weapon.angle +. Random.float desc.velocity_vary

    val mutable pos = 
      let px, py = params.Weapon.pos in
      float_of_int px, float_of_int py

    (* how long time the bullet already exists *)
    val mutable life_step = 0.
    
    (* we still have to move d_left during this frame *)
    val mutable d_left = 0., 0.
    
    val mutable deleted = false

    initializer
      let set = Sprite.Set.create 1 in
      Arena.set_boundary arena set;
      Sprite.Set.add set sprite

    method delete = 
      if not deleted then begin 
        deleted <- true;
        (* there's no need to delete the sprite, it has it's own set *)
        super#delete
      end

    (* move the ball by d_left *)
    method private continue_flight = 
      (* move the ball by d_left, but at most by step in each direction *)
      let rec move_by step = 
        let dx0, dy0 = d_left in
        let dx, dy = dx0, dy0 in
        (* decrease the distance to at most step (apropiately scaling the second coordinate) *)
        let ndx = min step (max (-.step) dx) in
        let dx, dy = if dx = ndx then dx, dy else ndx, dy /. dx *. ndx in 
        let ndy = min step (max (-.step) dy) in
        let dx, dy = if dy = ndy then dx, dy else dx /. dy *. ndy, ndy in
        (* move to the new position *)
        let x, y = pos in 
        let x, y = x +. dx, y +. dy in
        Sprite.move_to sprite (to_int (x, y));
        (* check if collides with player *)
        let list = Player.collides_with players sprite in
        let only_me = match list, params.Weapon.player with 
            [x], Some pl -> x == pl 
          | _ -> false 
        in
        if list <> [] && not (initial && only_me) then begin
          d_left <- 0., 0.;
          self#player_collision list
        end;
        (* check if move possible (on board and no collisions with arena) *)
        if Sprite.in_bounds sprite && 
          Sprite.collides_with sprite true (Arena.get_sprite_set arena) = [] 
        then begin
          pos <- (x, y);
            d_left <- dx0 -. dx, dy0 -. dy;
        end else
          if step <= 1. then begin
            (* farther flight impossible *)
            d_left <- dx0 -. dx, dy0 -. dy;
            self#rock_collision
          end else
            (* try shorter distance, the move can be possible *)
            move_by (step /. 2.)
      in
      move_by max_step;
      let dx, dy = d_left in
      if not deleted && (dx <> 0. || dy <> 0.) then
      self#continue_flight

    method process = begin
      if life_step >= Constants.crosshair_dist /. desc.initial_vel then
        initial <- false;
      let vx, vy = vel in
      let vx = vx *. (1. -. !Arena.delay *. desc.air_viscosity) in
      let vy = vy *. (1. -. !Arena.delay *. desc.air_viscosity) +. 
        !Arena.delay *. desc.weight 
      in
      vel <- (vx, vy);
      let mult = if life_step = 0. then Random.float 1. else 1. in
      d_left <- (vx *. !Arena.delay *. mult, vy *. !Arena.delay *. mult);
      Sprite.set_animation_frame sprite self#get_frame;
      self#continue_flight;
      Sprite.move_to sprite (to_int pos);
      if not deleted && life_step >= desc.life_time then begin
        self#end_of_life;
        self#delete
      end;
      (* increase life time *)
      life_step <- life_step +. !Arena.delay;
    end

    method bounce = 
      initial <- false;
      let vx, vy = vel in
      let dxl, dyl = d_left in
      let vx, vy, dxl, dyl = 
        if not (Sprite.in_bounds sprite) then begin
          (* bouncing from boundary *)
          let x0, y0, x1, y1 = Arena.get_boundary arena in
          let x, y = to_int pos in
          let x, y = x - spr_w / 2, y - spr_h / 2 in
          let xm = min (x - x0) (x1 - x - spr_w) in
          let ym = min (y - y0) (y1 - y - spr_h) in
          (* bounce from the nearest bound *)
          let vx, dxl = if xm <= ym then -. vx, -. dxl else vx, dxl in
          let vy, dyl = if ym <= xm then -. vy, -. dyl else vy, dyl in
          vx, vy, dxl, dyl 
        end else begin
          (* bouncing from rocks *)
          (* find the average point where the collision is: xp, yp *)
          let points = Sprite.collision_points sprite (Arena.get_sprite_set arena) in
          let xp, yp = List.fold_left (fun (x, y) (ax, ay) -> ax + x, ay + y) (0, 0) points in
          let l = float_of_int (List.length points) in
          let xp = -. float_of_int xp +. (float_of_int (spr_w-1) *. l /. 2.) in
          let yp = -. float_of_int yp +. (float_of_int (spr_h-1) *. l /. 2.) in
          (* the (xp, yp) vector is normal to the surface - we'll bounce according to it *)
          let dlp2 = xp *. xp +. yp *. yp in
          if classify_float dlp2 = FP_zero then 0., 0., 0., 0. else begin
            (* bounce the whole velocity: *)
            let il_sk = xp *. vx +. yp *. vy in
            let xs, ys = xp *. il_sk /. dlp2, yp *. il_sk /. dlp2 in
            let vx, vy = vx -. 2. *. xs, vy -. 2. *. ys in
            (* and bounce velocity left in this step: *)
            let il_sk = xp *. dxl +. yp *. dyl in
            let xs, ys = xp *. il_sk /. dlp2, yp *. il_sk /. dlp2 in
            let dxl, dyl = dxl -. 2. *. xs, dyl -. 2. *. ys in
            vx, vy, dxl, dyl
          end
        end
      in
      assert (classify_float vx <> FP_nan);
      assert (classify_float vy <> FP_nan);
      assert (classify_float dxl <> FP_nan);
      assert (classify_float dyl <> FP_nan);
      vel <- vx *. (1. -. desc.bounce_res), vy *. (1. -. desc.bounce_res);
      d_left <- dxl *. (1. -. desc.bounce_res), dyl *. (1. -. desc.bounce_res)

    method explode force = 
      if not deleted then begin
        let desc = { Explosion.std_desc with 
          Explosion.pos = to_int pos;
          Explosion.force = force;
          Explosion.owner = params.Weapon.player;
        } in
        let a = new Explosion.t set desc in
        self#delete
      end

    method draw cntx = Sprite.draw ~on:cntx sprite

    method virtual rock_collision : unit
    method virtual player_collision : Player.t list -> unit
    method virtual end_of_life : unit
    method virtual get_frame : int
  end;;
