
(** Standard bullet (flying object) *)

(** Description of a bullet. *)
type desc = {
  weight : float; (** gravity force *)
  air_viscosity : float; (** air viscosity *)
  initial_vel : float; (** initial velocity *)
  velocity_vary : float; (** fow much velocity can differ from given value *)
  life_time : float; (** life time *)
  bounce_res : float; (** how much the velocity decreases when bouncing *)
  img : Video.t; (** the image *)
  no_frames : int; (** number of frames in the image *)
};;

(** Typical description of a bullet. *)
val std_desc : desc;;

(** The bullet. *)
class virtual t : Bullet.set -> desc -> Player.shoot_params -> object
  inherit Bullet.t
  
  (** How much time already passed from when the object was created. *)
  val mutable life_step : float
  
  (** Position of the bullet. *)
  val mutable pos : Types.fpos
  
  (** Velocity of the bullet. *)
  val mutable vel : Types.vel

  val sprite : unit Sprite.t
  
  (** Called when it's time for new frame. Moves the object according to all
      the rules. *)
  method process : unit

  (** Called when there is a need to draw the object on a context. Draws the
      bullet. *)
  method draw : Video.t -> unit
  
  (** Call [explode force] when the bullet should explode with force
      [force]. *)
  method explode : int -> unit
  
  (** Call when the object should bounce from the boundary or from the rock. 
      Can be called only when the sprite is outside the boundary or collides
      with a rock, which means that usually it will be called from
      [rock_collision] method. *)
  method bounce : unit
  
  (** Removes the object from the set in which it is. Can be securely called
      when the object is already removed. *) 
  method delete : unit

  (** Will be called when the bulled collides with a rock or a boundary. *)  
  method virtual rock_collision : unit
  
  (** Will be called when the bulled collides with a player (takes a list of
      colliding players). *)
  method virtual player_collision : Player.t list -> unit
  
  (** Will be called when the [life_time] passed and the bulled should be
      removed. It will be removed itselft, but in this method you can for
      example make an explosion. *)
  method virtual end_of_life : unit
  
  (** Called every time when drawing the object, should return the number of
      frame which should be drawed. *)  
  method virtual get_frame : int
end;;
