
type 'player shoot_params = {
  player : 'player option;
  pos : Types.pos;
  vel_mod : Types.vel;
  angle : Types.angle;
  dir : Types.dir;
};;

type 'player fire_fun = 'player shoot_params -> unit;;

type check_stop_fun = unit -> bool;;

type 'player start_fire_fun = 'player shoot_params -> check_stop_fun -> unit;;

type 'player normal_t = { (* discrete weapon *)
  fire_fun : 'player fire_fun;
  amount_of_bullets : int;
  wait_time : float; (* time between consecutive shoots *)
  mutable bullets_left : int; (* bullets in magazine *)
  mutable wait_left : float; (* time to next shot allowed (counts down) *)
};;

type 'player cont_t = { (* continuous weapon *)
  start_fire_fun : 'player start_fire_fun;
  max_time : float; (* shot time between reloading *)
  mutable shoot_left : float; (* shoot time left before reloading *)
  mutable working : bool; (* was the button pressed from previous "process" call to now *)
  mutable working2 : bool ref; (* was the button pressed between the two last "process" calls *)
};;

type 'player specific_t = Normal of 'player normal_t | Continuous of 'player cont_t;;

type 'player t = {
  specific : 'player specific_t; (* data for discrete or continuous weapon *)
  reloading_time : float;
  name : string;
  mutable reloading_left : float; (* time left to end of reloading *)
  mutable is_reloading : bool;
};;

let create fire_fun amount_of_bullets reloading_time wait_time name =
  {
    reloading_time = reloading_time;
    name = name;
    reloading_left = 0.;
    is_reloading = false;
    specific = Normal 
    {
      fire_fun = fire_fun;
      amount_of_bullets = amount_of_bullets;
      wait_time = wait_time;
      bullets_left = amount_of_bullets;
      wait_left = 0.;
    };
  }
;;

let create_cont start_fun max_time reloading_time name = 
  {
    reloading_time = reloading_time;
    name = name;
    reloading_left = 0.;
    is_reloading = false;
    specific = Continuous
    {
      start_fire_fun = start_fun;
      max_time = max_time;
      shoot_left = max_time;
      working = false;
      working2 = ref false;
    };
  }
;;

(* returns percentage ammo *)
let get_ammo100 w = 
  if w.is_reloading then
    int_of_float ((w.reloading_time -. w.reloading_left) *. 100. /. w.reloading_time)
  else match w.specific with
      Normal s -> s.bullets_left * 100 / s.amount_of_bullets
    | Continuous s -> int_of_float (s.shoot_left *. 100. /. s.max_time)
;;

let is_reloading w = w.is_reloading;;

(* check if the reloading ended, eventually allow shooting *)
let check_stop_reloading w = 
  if w.reloading_left <= 0. then begin
    w.is_reloading <- false;
    match w.specific with 
        Normal s -> s.bullets_left <- s.amount_of_bullets
      | Continuous s -> s.shoot_left <- s.max_time
  end
;;

(* check if there is ammo in magazine, eventually start reloading *)
let check_start_reloading w = 
  if not w.is_reloading && match w.specific with 
      Normal s -> s.bullets_left = 0 
    | Continuous s -> s.shoot_left <= 0.
  then begin
    w.is_reloading <- true;
    w.reloading_left <- w.reloading_time;
    check_stop_reloading w
  end
;;

let process w = 
  begin match w.specific with
      Normal s -> 
        if s.wait_left > 0. then 
          s.wait_left <- s.wait_left -. !Arena.delay;
    | Continuous s ->
        s.working2 := s.working;
        s.working <- false;
        if !(s.working2) then
          s.shoot_left <- s.shoot_left -. !Arena.delay;
  end;
  if w.is_reloading then begin
    w.reloading_left <- w.reloading_left -. !Arena.delay;
    check_stop_reloading w
  end
;;

let fire w params = 
  begin match w.specific with
      Normal s -> 
        if s.bullets_left > 0 && s.wait_left <= 0. then begin
          s.bullets_left <- s.bullets_left - 1;
          s.wait_left <- s.wait_time;
          s.fire_fun params
        end
    | Continuous s -> 
        if s.shoot_left > 0. then begin 
          s.working <- true;
          if not !(s.working2) then begin
            (* starting *)
            let check = ref true in
            s.working2 <- check;
            s.start_fire_fun params (fun () -> not !check);
          end
        end
  end;
  check_start_reloading w;
;;

let get_name w = w.name;;


